% In this tutorial, we show how to use the Experiments package, which
% allows you to specify the evaluations you want to do and easily plot
% them. We first need to take a test script and slightly modify it. The
% test script is taken from the REPSLearner tutorial. The modifications are
% indicated by %!!. Having such a script, we can define a second script
% which defines the parameters that we want to evaluate and the number of
% trials. 

%!! Do not call any clear functionality. This is done by the getTrialForScript
%function. The function returns: 
% - a trial object
% - a settings object
% If you execute the script locally, then both of them are empty. If it is
% started by an evaluation, the settingsCluster contain the settings you should
% apply for your script. In the trial object, you should store all
% evaluation data you need.
[trial, settingsCluster] = Experiments.getTrialForScript(); 

%!! the trial object has 3 flags:
% -isDebug: execute locally... used for output and plotting
% -isConfigure: create all needed objects in the workspace. The whole
% workspace is then stored in the trial file automatically
% - isStart: run the evaluation
% If the trial is run locally, all 3 flags are set. Otherwise, isDebug is
% false and the script is called 2 times for each trial. At the trial
% creation, the isConfigure flag is set to true and all the objects are
% created. Subsequently, when the cluster job is started, the script is
% called with isStart set to true and isConfigure to false.


if (trial.isConfigure)
    % Copy paste from REPSLearner tutorial... in this tutorial we show how to use episodic REPS in combination with a
    % DMP based policy. Learning is only done for 1 iteration. 
    % For further details on DMPs, please refer to the TrajectoryGenerator tutorial.

    %!! set all default parameters for the experiment
    settings = Common.Settings();

    % Properties for the DMPs
    % we want to learn the temporal scaling factor -> 1 parameter
    settings.setProperty('useTau', 0);
    % we want to learn final angular positions of joints
    settings.setProperty('useGoalPos',1);
    % set numBasis to 5 (per degree of freedom)
    settings.setProperty('numBasis', 5);
    % we want to learn the weights of the DMP -> numBasis * numJoints
    % parameters
    settings.setProperty('useWeights', 1);
    % we end up with 16 parameters
    
    % Number of steps for an episode
    settings.setProperty('numTimeSteps', 100);
    settings.setProperty('numJoints',5);
    settings.setProperty('lambda',20);
    settings.setProperty('L',20);


    
    settings.setProperty('numSamplesEpisodes', 20);
    settings.setProperty('maxSamples', 20);           
    settings.setProperty('numIterations', 500);
    settings.setProperty('useImportanceWeighting', false);
    settings.setProperty('numInitialSamplesEpisodes',100);
    % Parameters of the InitialStateSampler
    settings.setProperty('InitialStateDistributionMinRange', - [0.5 zeros(1,2*settings.numJoints-1)]);
    settings.setProperty('InitialStateDistributionMaxRange', [0.5 zeros(1,2*settings.numJoints-1)]);
    settings.setProperty('InitialStateDistributionType', 'Uniform');
    
    %!! apply settings from the evaluation
    trial.configure(settingsCluster);
    

    % Creating an EpisodeWithStepsSampler.
    sampler = Sampler.EpisodeWithStepsSampler();

    dataManager = sampler.getEpisodeDataManager();

    % Set the isActiveSampler as an IsActiveNumSteps sampler, that always
    % terminates the episodes after 'numTimeSteps' steps. 
    sampler.getStepSampler.setIsActiveSampler(Sampler.IsActiveStepSampler.IsActiveNumSteps(dataManager));



    %% Define task setup, we use linear dynamics with numJoints = 3 joints
    environment = Environments.DynamicalSystems.LinearSystem(sampler, settings.numJoints);
    % Use the InitialStateSamplerStandard class to create a random starting 
    % position defined by the settings.
    initialStateSampler = Sampler.InitialSampler.InitialStateSamplerStandard(sampler);

    % register both in the sampler
    sampler.setTransitionFunction(environment);
    sampler.setInitialStateSampler(initialStateSampler);

    % Trajectory Generator
    dmp = TrajectoryGenerators.DynamicMovementPrimitives(dataManager,settings.numJoints);
    % Policy for tracking the trajectory
    controller = TrajectoryGenerators.TrajectoryTracker.LinearTrajectoryTracker(dataManager, settings.numJoints);
    % Register the linear PD trajectory tracker as action policy
    sampler.setActionPolicy(controller);

    % Reward Function: Our reward function is a via-point reward function that
    % punishes the distance to given via-points a3000t specific points in time. Our
    % via-points are defined in task space of the robot. 
    % To define the properties of our reward function, we use the struct viaPoint
    % viaPoint.times shows in which step we want to reach each point. In this
    % example, we want to reach point 1 after 40 steps and point 2 after 100 steps.
    viaPoint.times   = [100]; 
    % viaPoint.factors are the factors used to compute the costs for each dimension
    % of the state space. We use the factors <tt>[1000, 1000, 0, 0]</tt>, which
    % means that we only punish the distance of the end-effector position, but neglect 
    % the velocities. 
    viaPoint.factors = repmat([1e4, 1e4, 0, 0], length(viaPoint.times), 1);
    % The coordinates of the points we the robot should reach. 
     
    viaPoint.reachingStateOnWall{1} = [2,0, 0.0, 0.0 ];
    viaPoint.holeRadius = 0.15;
    %viaPoint.points{1}  = [1.0, 1.0, 0.0, 0.0];
    %viaPoint.points{2}  = [settings.numJoints, 0.0, 0.0, 0.0];
    % quadratic action cost factor for each time step
    viaPoint.uFactor = 10;

    % planarKinematics are used in the reward function to compute the
    % position of the end effector 
    planarKinematics = Environments.Misc.PlanarForwardKinematics(dataManager, settings.numJoints);

    % TaskSpaceViaPointRewardFunction uses the input form the viaPoint struct to calculate the reward
    
    rewardFunction = RewardFunctions.TimeDependent.TaskSpaceWallHoleRewardFunction(dataManager,planarKinematics, viaPoint.times,viaPoint.holeRadius,viaPoint.reachingStateOnWall,viaPoint.factors,viaPoint.uFactor);
                
    %rewardFunction = RewardFunctions.TimeDependent.TaskSpaceViaPointRewardFunction(dataManager, planarKinematics, viaPoint.times,viaPoint.points,viaPoint.factors,viaPoint.uFactor);

    % Return sampler computes the return. In this case we use the summed
    % rewards for each episode
    returnSampler = RewardFunctions.ReturnForEpisode.ReturnSummedReward(dataManager);

    % register reward and return function etc
    sampler.setRewardFunction(rewardFunction);
    sampler.setReturnFunction(returnSampler);

    % For the trajectory generators, we need to add the phaseGenerator, the
    % basisGenerator and the dmp's itself to the sampling pool
    % 'ParameterPolicy'. The addSamplerFunctionToPool method has the name of
    % the data manipulation function as its 2nd argument. These sampling
    % functions are called after the parameter policy (see below) has sampled a
    % new parameter vector. The phase, basis and reference trajectory depend on
    % the parameter, and, hence, will be different for each episode.
    sampler.addSamplerFunctionToPool('ParameterPolicy', 'generatePhase', dmp.phaseGenerator );
    sampler.addSamplerFunctionToPool('ParameterPolicy', 'generateBasis', dmp.basisGenerator);
    sampler.addSamplerFunctionToPool('ParameterPolicy', 'getReferenceTrajectory', dmp);

    %% Define Learning Algorithm
    % Parameter Policy. We want to learn a Gaussian policy in the parameters
    % that serves as the search distribution for REPS. This policy will be
    % registered as parameter policy and, hence, is used for sampling the
    % parameters of each episode.
    settings.setProperty('entropyBetaDiscount',0.4);
    settings.setProperty('epsilonAction',2);
    settings.setProperty('minEntropy',-5000);
    settings.setProperty('entropyUpdateType','absolute');

    
    parameterPolicy = Distributions.Gaussian.GaussianParameterPolicy(dataManager);
    sampler.setParameterPolicy(parameterPolicy);

    % REPS uses a (weighted) maximum likelihood update for the policy update.
    % In the case of a Gaussian distribution, this functionality is implemented
    % by the LinearGaussianMLLearner class. This class extracts input data, 
    % output data and the weights from the data set and computes a weighted ML
    % update for the mean and the covariance of the Gaussian distribution. 

    parameterPolicyLearner = Learner.SupervisedLearner.LinearGaussianMLLearner(dataManager, parameterPolicy);

    % The learner object will be used by the repsLearner to update the policy.
    % The REPS learner computes a weight for each data points and stores it in
    % the data structure (the weight name is given to the constructor, in our
    % case 'returnWeightings'). It also tells the parameterPolicyLearner object
    % to use the weights computed by REPS. After computing the weights, it
    % calls the updateModel function of the parameterPolicyLearner, which
    % performs the policy update with the weights stored in the data structure.
    % The parameterPolicyLearner can be any type of algorithm that performs a
    % weighted ML update. 
    %repsLearner = Learner.EpisodicRL.EpisodicREPS(dataManager, parameterPolicyLearner, 'returns', 'returnWeightings');
    
    policyLearner = Learner.EpisodicRL.CMAES2(dataManager, parameterPolicy); 
    policyLearner.initObject();

    %% Finalize and Init
    dataManager.finalizeDataManager();

    parameterPolicy.initObject();
   % PolicyLearner.initObject();

    %strategy for deleting old samples 
    deletionStrategy = LearningScenario.MaxSamplesDeletionStrategy();
    

    % The EpisodewithStepsSampler will now run 100 episodes for 100 steps.
    % Sampling is done in parallele for all 100 episodes.

    sampler.setParallelSampling(true);
end

%!! This flag is set if we want to execute the evaluation
if (trial.isStart)
    newData = dataManager.getDataObject(0);
    fullData = dataManager.getDataObject(0);
   
    for i = 1:settings.numIterations
        fullData.setDisableSubLayers(true);

        sampler.setSamplerIteration(i);


        fprintf('Generating Data\n');
        tic
        sampler.createSamples(newData);
        toc

        fullData.mergeData(newData);
        deletionStrategy.deleteSamples(fullData);


        fprintf('Iteration %d, Average Reward: %f\n', i, mean(newData.getDataEntry('returns')));

        tic
        policyLearner.updateModel(fullData);
        toc

        %!! Store some data. This will be stored in the data.mat file of
        %each trial
        trial.store('avgReturn', mean(newData.getDataEntry('returns')), Experiments.StoringType.ACCUMULATE);
        trial.store('muPolicy', parameterPolicy.getMean()', Experiments.StoringType.ACCUMULATE);
        %  trial.store('divMean', policyLearner.divMean, Experiments.StoringType.ACCUMULATE);
        %trial.store('divCov', policyLearner.divCov, Experiments.StoringType.ACCUMULATE);
        %!! Store the workspace (not necessary, but nice for debug)
       % trial.saveWorkspace();
        %!! store the trial (also not necessary at each iteration). Its
        %stored in a specific file specified by the evaluation
      %  trial.storeTrial();
        
               % fprintf('divMean: %f divCov: %f \n',trial.divMean(end) , trial.divCov(end));

    end    
end