% In this tutorial, we show how to use the Experiments package, which
% allows you to specify the evaluations you want to do and easily plot
% them. We first need to take a test script and slightly modify it. The
% test script is taken from the REPSLearner tutorial. The modifications are
% indicated by %!!. Having such a script, we can define a second script
% which defines the parameters that we want to evaluate and the number of
% trials. 

%!! Do not call any clear functionality. This is done by the getTrialForScript
%function. The function returns: 
% - a trial object
% - a settings object
% If you execute the script locally, then both of them are empty. If it is
% started by an evaluation, the settingsCluster contain the settings you should
% apply for your script. In the trial object, you should store all
% evaluation data you need.
[trial, settingsCluster] = Experiments.getTrialForScript(); 

%!! the trial object has 3 flags:
% -isDebug: execute locally... used for output and plotting
% -isConfigure: create all needed objects in the workspace. The whole
% workspace is then stored in the trial file automatically
% - isStart: run the evaluation
% If the trial is run locally, all 3 flags are set. Otherwise, isDebug is
% false and the script is called 2 times for each trial. At the trial
% creation, the isConfigure flag is set to true and all the objects are
% created. Subsequently, when the cluster job is started, the script is
% called with isStart set to true and isConfigure to false.


if (trial.isConfigure)

    settings = Common.Settings();

    settings.setProperty('lambda',15);
    settings.setProperty('L',15);


    settings.setProperty('numParameters', 15);
    settings.setProperty('numSamplesEpisodes', 15);
    settings.setProperty('maxSamples', 200);           
    settings.setProperty('numIterations', 1000);
    settings.setProperty('useImportanceWeighting', false);
    settings.setProperty('numInitialSamplesEpisodes',100);
    
    settings.setProperty('entropyBetaDiscount',0.15);
    settings.setProperty('epsilonAction',3);
    settings.setProperty('minEntropy',-5000);
    settings.setProperty('entropyUpdateType','absolute');
    
    %!! apply settings from the evaluation
    trial.configure(settingsCluster);
    

    % Creating an EpisodeWithStepsSampler.
   % sampler = Sampler.EpisodeWithStepsSampler();
    sampler = Sampler.EpisodeSampler();
    dataManager = sampler.getEpisodeDataManager();

    % Set the isActiveSampler as an IsActiveNumSteps sampler, that always
    % terminates the episodes after 'numTimeSteps' steps. 
    %sampler.getStepSampler.setIsActiveSampler(Sampler.IsActiveStepSampler.IsActiveNumSteps(dataManager));


    returnSampler = Environments.BanditEnvironments.RosenbrockReward(sampler, 0, settings.numParameters);

    sampler.setReturnFunction(returnSampler);
   

    %% Define Learning Algorithm
    % Parameter Policy. We want to learn a Gaussian policy in the parameters
    % that serves as the search distribution for REPS. This policy will be
    % registered as parameter policy and, hence, is used for sampling the
    % parameters of each episode.
    parameterPolicy = Distributions.Gaussian.GaussianParameterPolicy(dataManager);
    sampler.setParameterPolicy(parameterPolicy);

    % REPS uses a (weighted) maximum likelihood update for the policy update.
    % In the case of a Gaussian distribution, this functionality is implemented
    % by the LinearGaussianMLLearner class. This class extracts input data, 
    % output data and the weights from the data set and computes a weighted ML
    % update for the mean and the covariance of the Gaussian distribution. 

    parameterPolicyLearner = Learner.SupervisedLearner.LinearGaussianMLLearner(dataManager, parameterPolicy);

    % The learner object will be used by the repsLearner to update the policy.
    % The REPS learner computes a weight for each data points and stores it in
    % the data structure (the weight name is given to the constructor, in our
    % case 'returnWeightings'). It also tells the parameterPolicyLearner object
    % to use the weights computed by REPS. After computing the weights, it
    % calls the updateModel function of the parameterPolicyLearner, which
    % performs the policy update with the weights stored in the data structure.
    % The parameterPolicyLearner can be any type of algorithm that performs a
    % weighted ML update. 
    %repsLearner = Learner.EpisodicRL.EpisodicREPS(dataManager, parameterPolicyLearner, 'returns', 'returnWeightings');
    quadraticModel = Functions.SquaredFunction(dataManager, 'returns', 'parameters', 'rewardModel');
    quadraticModelLearner = Learner.ModelLearner.ConvexQuadraticModelLearner(dataManager, quadraticModel);
    importanceWeighting = DataPreprocessors.ImportanceSamplingLastKPolicies(dataManager, parameterPolicy);
    
    if (settings.useImportanceWeighting)
        quadraticModelLearner.setWeightName(importanceWeighting.importanceWeightName);
    end    
    
    policyLearner = Learner.EpisodicRL.ConstraintEntropyREPSCLosedForm(dataManager, parameterPolicy, quadraticModel, quadraticModelLearner, [], []);
   %policyLearner = Learner.EpisodicRL.NESLearner2(dataManager, parameterPolicy); 
   policyLearner.initObject();

    %% Finalize and Init
    dataManager.finalizeDataManager();

    parameterPolicy.initObject();
   % PolicyLearner.initObject();

    %strategy for deleting old samples 
    deletionStrategy = LearningScenario.MaxSamplesDeletionStrategy();
    

    % The EpisodewithStepsSampler will now run 100 episodes for 100 steps.
    % Sampling is done in parallele for all 100 episodes.

    sampler.setParallelSampling(true);
end

%!! This flag is set if we want to execute the evaluation
if (trial.isStart)
    newData = dataManager.getDataObject(0);
    fullData = dataManager.getDataObject(0);
   
    for i = 1:settings.numIterations
 

        sampler.setSamplerIteration(i);


        fprintf('Generating Data\n');
        tic
        sampler.createSamples(newData);
        toc

        fullData.mergeData(newData);
        deletionStrategy.deleteSamples(fullData);


        fprintf('Iteration %d, Average Reward: %f\n', i, mean(newData.getDataEntry('returns')));

        tic
        policyLearner.updateModel(fullData);
        toc

        %!! Store some data. This will be stored in the data.mat file of
        %each trial
        trial.store('avgReturn', mean(newData.getDataEntry('returns')), Experiments.StoringType.ACCUMULATE);
        trial.store('muPolicy', parameterPolicy.getMean()', Experiments.StoringType.ACCUMULATE);
        
        %!! Store the workspace (not necessary, but nice for debug)
        trial.saveWorkspace();
        %!! store the trial (also not necessary at each iteration). Its
        %stored in a specific file specified by the evaluation
        trial.storeTrial();
    end    
end