classdef EpisodicREPS < Learner.WeightedML.RLByWeightedML
    properties
        eta
        theta
        divKL
        stateFeatureDifference
        
        useHessian = true;
        useGradient = 'on';
        
        optimizationOutput
        
        numFeatures;
        
        stateFeatureName = '';
        
        sampleWeightingName = '';
        
        optimizer;
    end
    
    properties (SetObservable,AbortSet)
        epsilonAction = 0.5;
        numOptimizationsDualFunction = 300;
        alphaL2ThetaPunishment = 0.0;
        
        toleranceX = 10^-8;
        toleranceF = 10^-12;
        
        isDebug = false;
        
    end
    
    methods (Static)
        function [learner] = CreateFromTrial(trial)
            if (trial.isProperty('contextFeatures') && ~isempty( trial.contextFeatures))
                learner = Learner.EpisodicRL.EpisodicREPS(trial.dataManager, trial.parameterPolicyLearner, 'returns', 'returnWeightings', trial.contextFeatures.getFeatureName());
            else
                learner = Learner.EpisodicRL.EpisodicREPS(trial.dataManager, trial.parameterPolicyLearner, 'returns', 'returnWeightings');
            end
        end
        
        function [learner] = CreateFromTrialActionPolicy(trial)
            if (trial.isProperty('contextFeatures') && ~isempty( trial.contextFeatures))
                learner = Learner.EpisodicRL.EpisodicREPS(trial.dataManager, trial.policyLearner, 'returns', 'returnWeightings', trial.contextFeatures.getFeatureName());
            else
                learner = Learner.EpisodicRL.EpisodicREPS(trial.dataManager, trial.policyLearner, 'returns', 'returnWeightings');
            end
        end
        
    end
    
    % Class methods
    methods
        function obj = EpisodicREPS(dataManager, policyLearner, rewardName, weightName, stateFeatureName, varargin)
            
            if (numel(varargin) > 0)
                addArgument{1} = varargin{1};
            else
                addArgument = {};
            end
            
            obj = obj@Learner.WeightedML.RLByWeightedML(dataManager, policyLearner, rewardName, weightName, addArgument{:});
            
            if (~exist('stateFeatureName', 'var'))
                obj.stateFeatureName = '';
            else
                obj.stateFeatureName = stateFeatureName;
            end
            
            obj.linkProperty('epsilonAction');
            obj.linkProperty('alphaL2ThetaPunishment');
            
            obj.linkProperty('numOptimizationsDualFunction', 'maxNumOptiIterationsREPSDual');
            
            obj.linkProperty('toleranceF','REPSDualOptiAbsfTol');
            obj.linkProperty('toleranceX','REPSDualOptiAbsxTol');
            
            obj.linkProperty('numOptimizationsDualFunction');
            obj.linkProperty('toleranceX', 'REPSDualOptimizationToleranceX');
            obj.linkProperty('toleranceF', 'REPSDualOptimizationToleranceF');
            
            obj.linkProperty('isDebug', 'DebugMode');
            
            if (size(varargin,2)>1)
                obj.setSampleWeightingName(varargin{2});
            else
                obj.registerWeightingFunction();
            end
        end
        
        function [] = setSampleWeightingName(obj, name)
            obj.sampleWeightingName = name;
            obj.registerWeightingFunction();
        end
        
        function [] = initObject(obj)
            obj.theta = zeros(obj.dataManager.getNumDimensions(obj.stateFeatureName), 1);
            obj.eta = 1;
            obj.iteration = 1;
            obj.numFeatures = obj.dataManager.getNumDimensions(obj.stateFeatureName);
            [params, lowerBound, upperBound] = obj.getParametersForOptimization();
            obj.optimizer = Optimizer.OptimizerFactory.createOptimizer(obj.getNumDualParams(), 'EpisodicREPSOptimization', ...
                lowerBound, upperBound);
        end
        
        function [] = setStateFeatureName(obj, stateFeatureName)
            obj.stateFeatureName = stateFeatureName;
            obj.registerWeightingFunction();
        end
        
        function [returnweighting] = computeWeighting(obj, varargin) %input arguments are as follows, rewards, stateFeatures, localityWeighting
            
            repsData = getREPSDataStructure(obj, varargin{:});
            if (obj.eta < std(repsData.rewards) * 0.1)
                obj.eta = std(repsData.rewards) * 0.1;
            end
            obj.divKL = 100;
            
            returnweighting = obj.computeWeightingFromEtaAndTheta(repsData, obj.eta, obj.theta);
            returnweighting = returnweighting / sum(returnweighting);
            obj.divKL = obj.getKLDivergence(repsData.sampleWeighting, returnweighting);
            obj.stateFeatureDifference = repsData.meanStateFeatures - sum(bsxfun(@times, repsData.stateFeatures, returnweighting), 1);
                        
            bestDiv = Inf;
            lastDiv = Inf;
            withoutImprovement=0;
            for i = 1:40
                [params] = obj.optimizeDualFunction(repsData, obj.getOptimizationOptions());
                
                obj.setParametersFromOptimization(params);
                
                weighting = obj.computeWeightingFromEtaAndTheta(repsData, obj.eta, obj.theta);
                weighting = weighting / sum(weighting);
                divKL = obj.getKLDivergence(repsData.sampleWeighting, weighting);
                
                
                if (divKL > 3 || isnan(divKL))
                    disp('diVKL warning');
                    %keyboard;
                end
                stateFeatureDifference = repsData.meanStateFeatures - sum(bsxfun(@times, repsData.stateFeatures, weighting), 1);
                
                fprintf('Feature Error: %f, KL: %f\n', full(max(abs(stateFeatureDifference))), divKL);
                
                if (bestDiv ~= Inf && i>=10 && max(abs(stateFeatureDifference)) >= bestDiv)
                    withoutImprovement = withoutImprovement+1;
                end
                if withoutImprovement>=3
                    disp('No improvement within the last 3 iterations.');
                    break;
                end
                %if (abs(divKL - obj.epsilonAction) < 0.05 && (isempty(stateFeatureDifference) || max(abs(stateFeatureDifference)) < 0.001 ))
                if (abs(divKL - obj.epsilonAction) < 0.05 && (isempty(stateFeatureDifference) || (max(abs(stateFeatureDifference)) < 0.01  && max(abs(stateFeatureDifference)) < bestDiv)))
                    disp('Accepted solution.');
                    withoutImprovement = 0;
                    obj.divKL = divKL;
                    obj.stateFeatureDifference = stateFeatureDifference;
                    returnweighting = weighting;
                    
                    bestDiv = max(abs(stateFeatureDifference));
                    
                    if (abs(divKL - obj.epsilonAction) < 0.05 && (isempty(stateFeatureDifference) || max(abs(stateFeatureDifference)) < 0.001 ))
                        disp('Found sufficient solution.');
                        break;
                    end
                end
                
                if(full(max(abs(stateFeatureDifference)-lastDiv))>-0.000001)
                    disp('Solution unchanged or degrading, restart from new point');
                    obj.theta = rand(size(obj.theta))*2-1;
                    lastDiv = Inf;
                else
                    lastDiv = max(abs(stateFeatureDifference));
                end
            end
        end
        
        function [] = printMessage(obj, data)
            fprintf('Eta: %f, divKL %f, stateFeatureError: %f\n', obj.eta, obj.divKL, full(max(abs(obj.stateFeatureDifference))));
        end
        
        function [] = addDefaultCriteria(obj, trial, evaluationCriterion)
            evaluationCriterion.addCriterion('endLoop', '', 'divKL', Experiments.StoringType.ACCUMULATE, @()obj.divKL);
        end
    end
    
    methods (Access = protected)
        function obj = setParametersFromOptimization(obj, params)
            obj.theta = params(1:obj.numFeatures );
            obj.eta = params(1 + obj.numFeatures);
        end
        
        function numParams = getNumDualParams(obj)
            numParams = obj.numFeatures + 1;
        end
        
        function [params, lowerBound, upperBound] = getParametersForOptimization(obj)
            params = [obj.theta; obj.eta];
            %params = [ones(size(obj.theta)); obj.eta];
            
            lowerBound = [- ones(obj.numFeatures, 1) * 1e10;  1e-20];
            upperBound = [ones(obj.numFeatures, 1) * 1e10; 1e10];
        end
        
        function [repsData] = getREPSDataStructure(obj, rewards, stateFeatures, sampleWeighting)
            if (~exist('stateFeatures', 'var'))
                stateFeatures = [];
            end
            
            if (~exist('sampleWeighting', 'var') || sum(sampleWeighting)==0)
                sampleWeighting = ones(size(rewards,1),1);
            end
            % rewardMean = mean(rewards);
            repsData.rewards = rewards;
            repsData.stateFeatures = stateFeatures;
            repsData.sampleWeighting = sampleWeighting;
            
            repsData.meanStateFeatures = sum(bsxfun(@times, repsData.stateFeatures, repsData.sampleWeighting / sum(repsData.sampleWeighting)));
        end
        
        
        
        function [newParams,returnCode] = optimizeDualFunction(obj, repsData, optimizationOptions)
            
            [params, lowerBound, upperBound] = obj.getParametersForOptimization();
            
            if (obj.isDebug && obj.iteration == 1)
                [g, gD, gDNumeric] = obj.numericDualFunctionGradient(obj.getParametersForOptimization(), repsData);
                fprintf('Gradient Check: MAE of numeric Gradient %f\n', sum(abs(gD - gDNumeric)));
            end
            
            %options = optimizationOptions;
            
            dualFunctionAnonymous = @(params) dualFunction(obj, params, repsData);
            
            
            %DIRTY! Wait for license block
            %while (~license('checkout','Optimization_Toolbox'))
            %    pause(30);
            %end
            [newParams, val, returnCode] = obj.optimizer.optimize(dualFunctionAnonymous, params);
            %[newParams,fval,exitflag,obj.optimizationOutput] = fmincon(dualFunctionAnonymous,params,[],[],[],[],lowerBound,upperBound, [], options);
            
            %optimizer = Optimizer.
        end
        
        function [g, gD] = dualTheta(obj, params, repsData)
            params  = [params; obj.eta;];
            [g, gD] = obj.dualFunction(params, repsData);
            gD      = gD(1:obj.numFeatures);
        end
        
        %%
        function [g, gD] = dualEta(obj, params, repsData)
            params  = [obj.theta   ; params];
            [g, gD] = obj.dualFunction(params, repsData);
            gD      = gD(end);
        end
        
        %%
        function [g, gD] = dualFunction(obj, params, repsData)
            
            numFeatures = obj.numFeatures;
            theta       = params(1:numFeatures);
            eta         = params(numFeatures + 1);
            
            phiMat      = repsData.stateFeatures;
            phiHat      = repsData.meanStateFeatures;
            
            R           = repsData.rewards;
            epsilon     = obj.epsilonAction;
            
            if (~isempty(theta))
                
                V       = phiMat * theta;
                VHat    = phiHat * theta;
            else
                V       = 0;
                VHat    = 0;
            end
            importanceWeights           = repsData.sampleWeighting;
            
            advantage                   = (R - V);
            [maxAdvantage, ~]           = max(advantage);
            R                           = R - maxAdvantage;
            advantage                   = (R - V) / eta;
            boundedAdvantage            = advantage;
            
            g                           = 0;
            gD                          = zeros(size(params));
            
            if( max(boundedAdvantage) > 500 )
                g           = 1e30 - eta;
                gD(end-1)   = -1;
                return
            end
            
            N                       = sum(importanceWeights);
            expAdvantage            = importanceWeights .* exp(boundedAdvantage);
            sumExpAdvantage         = sum(expAdvantage);
            
            if(sumExpAdvantage < realmin )
                sumExpAdvantage = realmin;
            end
            if(N < realmin)
                N = realmin;
            end
            
            gLogPart                = 1 / N * sumExpAdvantage;
            
            g = g + eta*log(gLogPart) + VHat +  maxAdvantage;
            g = g + eta * epsilon + obj.alphaL2ThetaPunishment * sum(theta.^2);
            
            if (nargout > 1)
                gDEta = epsilon + log(gLogPart) + sum(expAdvantage .* (R - V)) / ...
                    (- eta * sumExpAdvantage);
                if (-eta *sumExpAdvantage ==0)
                    gDEta = 1e100;
                end
                gD(numFeatures + 1) = gDEta;
                
                if (~isempty(theta))
                    gDTheta = phiHat + sum(bsxfun(@times, -phiMat, expAdvantage)) /...
                        (sumExpAdvantage) + 2 * obj.alphaL2ThetaPunishment * theta';
                    gD(1:numFeatures) = gDTheta;
                end
            end
            
            if (isnan(g) || sum(isnan(gD)) > 0)
                if(evalin('base','exist(''manager'')') && ~evalin('base','manager')) %throw errors for workers to exit.
                    error('Dual Function');
                end
                
                %keyboard;
            end
        end
        
        %%
        function [weighting] = computeWeightingFromEtaAndTheta(obj, repsData, eta, theta)
            
            if (~isempty(theta))
                advantage = repsData.rewards - repsData.stateFeatures * obj.theta;
            else
                advantage = repsData.rewards;
            end
            
            weighting = repsData.sampleWeighting .* exp( (advantage - max(advantage)) / eta);
        end
        
        
        
        %%
        function [g, gD, gDNumeric] = numericDualFunctionGradient(obj, params, data)
            gDNumeric = zeros(length(params),1);
            [g,gD] = obj.dualFunction(params, data);
            stepSize = max(min(abs(params) * 1e-4, 1e-6), 1e-6);
            for i = 1:length(params)
                
                paramsTemp = params;
                paramsTemp(i) = params(i) - stepSize(i) ;
                
                g1 = obj.dualFunction(paramsTemp, data);
                
                paramsTemp = params;
                paramsTemp(i) = params(i) + stepSize(i);
                
                g2 = obj.dualFunction(paramsTemp, data);
                gDNumeric(i) = (g2 - g1) / (stepSize(i)* 2);
            end
        end
        
        function optimizationOptions = getOptimizationOptions(obj)
            
            display = 'off';
            if (obj.isDebug )
                display = 'iter';
            end
            
            optimizationOptions = optimset('GradObj',obj.useGradient, 'Display',display,'MaxFunEvals', obj.numOptimizationsDualFunction * 5,...
                'Algorithm', 'trust-region-reflective','TolX', obj.toleranceX, 'TolFun', obj.toleranceF, ...
                'MaxIter', obj.numOptimizationsDualFunction );
        end
        
        
    end
    
    methods (Access = protected)
        function [] = registerWeightingFunction(obj)
            if (~ isempty(obj.sampleWeightingName))
                obj.addDataManipulationFunction('computeWeighting', {obj.rewardName, obj.stateFeatureName, obj.sampleWeightingName, obj.additionalInputData{:}}, {obj.outputWeightName});
            else
                obj.addDataManipulationFunction('computeWeighting', {obj.rewardName, obj.stateFeatureName, obj.additionalInputData{:}}, {obj.outputWeightName});
            end
        end
        
    end
end

